"""
The monk 1 task is (f1=f2)or(f5=1)
This makes a two layer net, with spikesources for each line of inputs
and one per output line.
The layers are connected by STDP synapses.
It trains for NUMBER_TRAINING_EPOCHS.  
The output neurons give the right results 12 of the 31 f5=1 feature.
"""
import sys
import cPickle as pickle

import pyNN.nest as sim

TIME_BETWEEN_INPUT_OUTPUT = 5

EXAMPLE_TIME = 100.0
NUMBER_TRAINING_EPOCHS = 20
NUMBER_CATEGORIES = 2

TIME_STEP = 1.0
FIRING_WEIGHT = 0.1

NEURONS_PER_FEATURE = 1
NUMBER_INPUT_LINES = 124

RUN_DURATION = EXAMPLE_TIME*NUMBER_INPUT_LINES*NUMBER_TRAINING_EPOCHS

neuronsInFirstLayer = 17
neuronsInSecondLayer = NUMBER_CATEGORIES

firstSecond = -1 #global for printing out synaptic weights

#initialize the simulator. 
def init(): 
    sim.setup(timestep=TIME_STEP,min_delay=TIME_STEP,
                    max_delay=TIME_STEP, debug=0)

def readData(fileName,inputLines):
    data = []
    dataFileHandle = open(fileName,'r')
    for lineNumber in range (0,NUMBER_INPUT_LINES):
        inputLine = dataFileHandle.readline()
        dataItemStrings = inputLine.split()
        category = int(dataItemStrings[0])
        feature0 = int(dataItemStrings[1])
        feature1 = int(dataItemStrings[2])
        feature2 = int(dataItemStrings[3])
        feature3 = int(dataItemStrings[4])
        feature4 = int(dataItemStrings[5])
        feature5 = int(dataItemStrings[6])
        dataLine = [category,feature0,feature1,feature2,feature3,feature4,feature5]
        #print dataLine,dataItemStrings[7]
        data = data + [dataLine]

    dataFileHandle.close()
    return data

#Make spike sources to turn on the first layer, one for each data line.
def createInputSpikeSources(start,numberSources):
    epochTime = EXAMPLE_TIME*numberSources
    inputSpikeTimesMatrix = []

    for sourceNumber in range (0,numberSources):
        inputSpikeTimes = []
        for epoch in range (0,NUMBER_TRAINING_EPOCHS): 
            offset = sourceNumber * EXAMPLE_TIME
            inputSpikeTime = start+offset+ (epoch*epochTime)
            inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]

        inputSpikeTimesMatrix = inputSpikeTimesMatrix + [inputSpikeTimes]

    #print "inputSpikeTimes", inputSpikeTimesMatrix
    inputSpikeGens = sim.Population(numberSources,sim.SpikeSourceArray,
                {'spike_times': inputSpikeTimesMatrix},label="inputSpikes")

    return inputSpikeGens

#Make one spike source for output for each data line.
def createOutputSpikeSources(start,numberSources):
    #Create the output spike sources and when they fire
    epochTime = EXAMPLE_TIME*numberSources
    outputSpikeTimesMatrix = []
    learnOffset = TIME_BETWEEN_INPUT_OUTPUT
    for sourceNumber in range (0,numberSources):
        outputSpikeTimes = [] 
        for epoch in range (0,NUMBER_TRAINING_EPOCHS): 
            offset = (sourceNumber * EXAMPLE_TIME) + TIME_BETWEEN_INPUT_OUTPUT
            outputSpikeTime = start+offset+ (epoch*epochTime)
            outputSpikeTimes = outputSpikeTimes + [outputSpikeTime]

        outputSpikeTimesMatrix = outputSpikeTimesMatrix + [outputSpikeTimes]

    outputSpikeGens = sim.Population(numberSources,sim.SpikeSourceArray,
                {'spike_times': outputSpikeTimesMatrix},label="outputSpikes")

    #print "outputSpikeTimes", outputSpikeTimesMatrix
    return outputSpikeGens

#This just creates the sources.  One input and one output for
#each data line.
def createTrainSources(numberSources):
    firstSpike = 10
    inputSpikeGens = createInputSpikeSources (firstSpike,numberSources)
    outputSpikeGens = createOutputSpikeSources(firstSpike,numberSources)

    return [inputSpikeGens,outputSpikeGens]


def makeInputToFirstLayerSynapses(dataSourceNumber,dataItem):
    #dataItem[0] is the cat
    feature1Offset = (dataItem[1]-1)
    inputConnector =  [(dataSourceNumber,feature1Offset,FIRING_WEIGHT,TIME_STEP)]
    feature2Offset = (dataItem[2]-1) + 3 #There are 3 first features
    inputConnector =  inputConnector + [(dataSourceNumber,feature2Offset,FIRING_WEIGHT,TIME_STEP)]
    feature3Offset = (dataItem[3]-1) + 6 #There are 3 second features
    inputConnector =  inputConnector + [(dataSourceNumber,feature3Offset,FIRING_WEIGHT,TIME_STEP)]
    feature4Offset = (dataItem[4]-1) + 8 #There are 2 third features
    inputConnector =  inputConnector + [(dataSourceNumber,feature4Offset,FIRING_WEIGHT,TIME_STEP)]
    feature5Offset = (dataItem[5]-1) + 11 #There are 3 fourth features
    inputConnector =  inputConnector + [(dataSourceNumber,feature5Offset,FIRING_WEIGHT,TIME_STEP)]
    feature6Offset = (dataItem[6]-1) + 15 #There are 4 fifth features
    inputConnector =  inputConnector + [(dataSourceNumber,feature6Offset,FIRING_WEIGHT,TIME_STEP)]

    #print "source to input layer", inputConnector
    return inputConnector

def connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells,data):
    numberDataItems = NUMBER_INPUT_LINES
    for dataItem in range (0,numberDataItems):
        inputConnector= makeInputToFirstLayerSynapses(dataItem,data[dataItem])
        inputFromListConnector = sim.FromListConnector(inputConnector)
        #print "source to input",inputConnector
        sim.Projection(inputSpikeSources, inputCells, inputFromListConnector, 
                   receptor_type='excitatory')

def connectSpikeSourcesToSecondLayer(outputSpikeSources,outputCells,data):
    outputConnector = []
    for lineNumber in range (0,NUMBER_INPUT_LINES):
        category = (data[lineNumber][0])

        outputConnector = outputConnector + [(lineNumber,category,FIRING_WEIGHT,TIME_STEP)]
            
    #print outputConnector
    outputFromListConnector = sim.FromListConnector(outputConnector)
    sim.Projection(outputSpikeSources, outputCells,outputFromListConnector, 
                   receptor_type='excitatory')

def connectInput(spikeSources, inputCells,outputCells,data):
    inputSpikeSources = spikeSources[0]
    outputSpikeSources = spikeSources[1]

    connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells,data)
    connectSpikeSourcesToSecondLayer(outputSpikeSources,outputCells,data)


def connectFirstToSecondLayer(fromCells, toCells):
    global firstSecond
    connector = []
    stdp_model = sim.STDPMechanism(
        timing_dependence=sim.SpikePairRule(tau_plus=20.0,tau_minus=20.0,
                                            A_plus=0.005, A_minus=0.005),
        weight_dependence=sim.MultiplicativeWeightDependence(w_min=0.0, w_max=0.02),
        weight=0.0000,
        delay=1.0,
        dendritic_delay_fraction=float(1))

    connector = []
    for firstLayerNeuron in range (0,neuronsInFirstLayer):
        for secondLayerNeuron in range (0, neuronsInSecondLayer):
            weight = 0.0
            connector = connector + [(firstLayerNeuron,
                    secondLayerNeuron,weight,TIME_STEP)]
    fromListConnector = sim.FromListConnector(connector)

    firstSecond = sim.Projection(fromCells,toCells,
                                 fromListConnector,
                                 synapse_type = stdp_model)

def connectTestFirstToSecondLayer(fromCells, toCells, synapses):
    connector = []
    for synapseOffset in range (0,len(synapses)):
        fromNeuron = synapses[synapseOffset][0]
        toNeuron = synapses[synapseOffset][1]
        weight = synapses[synapseOffset][2]
        connector = connector + [(fromNeuron,toNeuron,
                        weight,TIME_STEP)]
    testFromListConnector = sim.FromListConnector(connector)
    sim.Projection(fromCells,toCells, testFromListConnector, 
                   receptor_type='excitatory')

def createDefaultNeurons(numberNeurons):
    cells=sim.Population(numberNeurons,sim.IF_cond_exp,cellparams={})
    return cells

def setupRecording(cells):
    cells.record({'spikes','v'})

def printResults(simCells,file):
    simCells.write_data(file+".pkl",'spikes')
    #simCells.write_data(file+"V.pkl",'v')
    #simCells.write_data(file+"W.pkl",'w')

def printF5Equal1(data):
    for lineNumber in range (0,NUMBER_INPUT_LINES):
        feature5 = data[lineNumber][5]
        if (feature5==1):
            print 62000+(lineNumber*50)

def printSTDPSynapses(fileName,synapseMatrix):
    #for synapse in synapseMatrix:
    #    print synapse[0],synapse[1],synapse[2]
    fileHandle = open(fileName,'w')
    for synapse in synapseMatrix:
        outString=str(synapse[0])+" "+ str(synapse[1])+" "+str(synapse[2])+"\n"
        fileHandle.write(outString)
    fileHandle.close()

simName = "nest"
spinnVersion = -1

init()
data=readData("monkData/monks-1.train",NUMBER_INPUT_LINES)
trainSpikeGenerators = createTrainSources(NUMBER_INPUT_LINES)

firstLayerCells = createDefaultNeurons(neuronsInFirstLayer)
secondLayerCells = createDefaultNeurons(neuronsInSecondLayer)

#hook it up
connectInput(trainSpikeGenerators,firstLayerCells,secondLayerCells,data)

#make the stdp synapses
connectFirstToSecondLayer(firstLayerCells,secondLayerCells)

setupRecording(firstLayerCells)
setupRecording(secondLayerCells)

print RUN_DURATION
sim.run(RUN_DURATION)

printResults(firstLayerCells,"output/first")
printResults(secondLayerCells,"output/second")

#printF5Equal1(data)

firstSecondSynapseWeights = firstSecond.get(["weight"],format="list")
printSTDPSynapses("output/oneFirstSecondSynapses.txt",firstSecondSynapseWeights)

#print "stdp synapse matrix ", firstSecondSynapseWeights


