import pandas as pd
import numpy as np
import string
import nltk
from gensim.models.word2vec import Word2Vec
from multiprocessing import cpu_count
from sklearn.model_selection import train_test_split
from gensim.models.tfidfmodel import TfidfModel
from gensim.corpora import Dictionary
import sys
def word2vec(input_data):
    model = Word2Vec(input_data,vector_size =100,window=8, min_count=1,workers=cpu_count())
    model.init_sims(replace=True)
    return model
def removeStopWords(stopWords, txt):
    newtxt = ' '.join([word for word in txt.split() if word not in stopWords])
    return newtxt
def get_tfidf_model(texts):
    dct = Dictionary(texts)  # fit dictionary
    corpus = [dct.doc2bow(line) for line in texts]  # convert corpus to BoW format
    model = TfidfModel(corpus)  # fit model
    return model,dct # apply model to the first corpus document
def get_tfidf(model,dic,text):
    corpus = dic.doc2bow(text)
    vector = model[corpus]
    return [value[1] for value in vector]

def text_to_vector(word2vec,tfidf,dct,vector_type,data,train_sample,test_sample):
    vector_type='MIX'
    #get the .groupby('Class Index')training sample
    #sample_train = data.groupby('Class Index').apply(lambda x: x.sample(train_sample,random_state = 22))
    sample_train = data.apply(lambda x: x.sample(n=train_sample,random_state = 22))
    sentences_train = [sent.split(" ") for sent in sample_train.Description]
    #get the testing sample
    #sample_test = data.groupby('Class Index').apply(lambda x: x.sample(test_sample))
    sample_test = data.apply(lambda x: x.sample(n=test_sample,random_state = 9))
    sentences_test= [sent.split(" ") for sent in sample_test.Description]

    #trainData = [[np.mean(model.wv[word]) for word in sent] for sent in sentences]
    X_train,X_test,y_train,y_test =sentences_train,sentences_test,sample_train['Class Index'],sample_test['Class Index']

    if vector_type =='AVG':
        train_vectors = [np.mean([word2vec.wv.word_vec(word) for word in sent],axis=0)for sent in X_train]
        test_vectors = [np.mean([word2vec.wv.word_vec(word) for word in sent],axis=0)for sent in X_test]
    if vector_type=='MIX':
        train_vectors = [np.mean([word2vec.wv.word_vec(word)*value for word,value in zip(sent,get_tfidf(tfidf,dct,sent))],axis=0)  for sent in X_train]
        test_vectors = [np.mean([word2vec.wv.word_vec(word)*value for word,value in zip(sent,get_tfidf(tfidf,dct,sent))],axis=0)  for sent in X_test]
    else:
        train_vectors = [np.mean([word2vec.wv.word_vec(word) for word in sent],axis=0)for sent in X_train]
        test_vectors = [np.mean([word2vec.wv.word_vec(word) for word in sent],axis=0)for sent in X_test]

    train_vectors = [[abs(int(value*1000)) for value in words]for words in train_vectors]
    test_vectors = [[abs(int(value*1000)) for value in words]for words in test_vectors]
    return train_vectors,test_vectors,y_train,y_test



def preprocessing(path,map_type):
    data = pd.read_csv(path)
    #transform words to lower
    data.Description = [w.lower() for w in data.Description]

    #remoove punctuation
    table = str.maketrans(dict.fromkeys(string.punctuation))  # OR {key: None for key in string.punctuation}
                                                               # Output: string without punctuation

    data.Description = [w.translate(table) for w in data.Description]
    #remove stopWords

    stops = nltk.corpus.stopwords.words("english")
    data.Description = [removeStopWords(stops,w) for w in data.Description]
    #make the data usable to train word2vec

    texts = [[text for text in doc.split()] for doc in data.Description]

    #get word2vec models
    model_word2vec = word2vec(texts)
    model_tfidf,dic= get_tfidf_model(texts)
    #get data
    X_train,X_test,y_train,y_test = text_to_vector(model_word2vec,model_tfidf,dic,map_type,data,train_sample=1000,test_sample = 200)
    #apply word
    return X_train,X_test,y_train,y_test

path = 'dataset/train.csv'

args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 2):
    type_processing = args[1]
else:
    type_processing = "AVG"
X_train,X_test,y_train,y_test = preprocessing(path,type_processing)
if type_processing == "AVG":
    np.save('processed_data/X_train', X_train, allow_pickle=True)
    np.save('processed_data/X_test', X_test, allow_pickle=True)
    np.save('processed_data/y_train', y_train, allow_pickle=True)
    np.save('processed_data/y_test', y_test, allow_pickle=True)
elif type_processing =="MIX":
    np.save('processed_data/X_train_tf', X_train, allow_pickle=True)
    np.save('processed_data/X_test_tf', X_test, allow_pickle=True)
    np.save('processed_data/y_train', y_train, allow_pickle=True)
    np.save('processed_data/y_test', y_test, allow_pickle=True)
