"""
The monk 1 task is (f1=f2)or(f5=1)
Calculate the training covariance between the feature pairs and the
categories.  The matrix has a column for each feature value pair. 
The rows go feature1, value1,feature2, value2, numCat0s,numCat1s,
percentageCat0s, percentageCat1

"""
import numpy as np

NUMBER_INPUT_LINES = 124
NUMBER_FEATURE_PAIRS = 119
NUMBER_FEATURE_VALUES = 17
FEATURES = 6
FEATURE_VALUES = [3,3,2,3,4,2]
REST_FEATURE_VALUES = [14,11,9,6,2,0]

def readData(fileName,inputLines):
    data = []
    dataFileHandle = open(fileName,'r')
    for lineNumber in range (0,NUMBER_INPUT_LINES):
        inputLine = dataFileHandle.readline()
        dataItemStrings = inputLine.split()
        category = int(dataItemStrings[0])
        feature0 = int(dataItemStrings[1])
        feature1 = int(dataItemStrings[2])
        feature2 = int(dataItemStrings[3])
        feature3 = int(dataItemStrings[4])
        feature4 = int(dataItemStrings[5])
        feature5 = int(dataItemStrings[6])
        dataLine = [category,feature0,feature1,feature2,feature3,feature4,feature5]
        #print dataLine,dataItemStrings[7]
        data = data + [dataLine]

    dataFileHandle.close()
    return data

def addLabel(firstFeature,startFeatureVal,matrix):
    restFeatureValues = REST_FEATURE_VALUES[firstFeature]
    for firstFeatureValue in range (0,FEATURE_VALUES[firstFeature]):
        currentOtherFeature = firstFeature+1
        otherFeatureValue = 0
        for restFeatureValue in range (0,restFeatureValues):
            matrixColumn = (firstFeatureValue*restFeatureValues)
            matrixColumn += startFeatureVal+restFeatureValue
            matrix[0,matrixColumn] = firstFeature+1
            matrix[1,matrixColumn] = firstFeatureValue+1
            matrix[2,matrixColumn] = currentOtherFeature+1
            matrix[3,matrixColumn] = otherFeatureValue+1
            otherFeatureValue += 1
            if (otherFeatureValue == FEATURE_VALUES[currentOtherFeature]):
                currentOtherFeature += 1
                otherFeatureValue = 0
            

def labelMatrix(matrix):
    addLabel(0,0,matrix)
    addLabel(1,42,matrix)
    addLabel(2,75,matrix)
    addLabel(3,93,matrix)
    addLabel(4,111,matrix)

def getFirstFeatureValuePair(matrixColumn):
    feature = -1
    value = -2

    if (matrixColumn < 42):
        feature = 1
        value = matrixColumn / REST_FEATURE_VALUES[0]
    elif (matrixColumn < 75):
        feature = 2
        offsetColumn = matrixColumn - 42
        value = offsetColumn / REST_FEATURE_VALUES[1]
    elif (matrixColumn < 93):
        feature = 3
        offsetColumn = matrixColumn - 75
        value = offsetColumn / REST_FEATURE_VALUES[2]
    elif (matrixColumn < 111):
        feature = 4
        offsetColumn = matrixColumn - 93
        value = offsetColumn / REST_FEATURE_VALUES[3]
    elif (matrixColumn < NUMBER_FEATURE_PAIRS):
        feature = 5
        offsetColumn = matrixColumn - 111
        value = offsetColumn / REST_FEATURE_VALUES[4]
    else:
        print "wrong input", matrixColumn

    value += 1

    return [feature,value]

def getSecondFeatureValuePair(matrixColumn,matrix):
    feature = matrix[2,matrixColumn]
    value = matrix[3,matrixColumn]

    return [feature,value]

def getColumn(feature1,value1,feature2,value2,matrix):
    for column in range (0,NUMBER_FEATURE_PAIRS):
        if ((matrix[0][column] == feature1) and (matrix[1][column] == value1)and
            (matrix[2][column] == feature2) and (matrix[3][column] == value2)):
            return column

def addPair (firstFeature,secondFeature,dataLine,matrix):
    category = dataLine[0]
    firstValue = dataLine[firstFeature]
    secondValue = dataLine[secondFeature]
    column = getColumn(firstFeature,firstValue,secondFeature,secondValue,matrix)
    if (category == 0):
        matrix[4][column] += 1
    elif (category == 1):
        matrix[5][column] += 1

    #print category,firstValue,secondValue


def fillMatrix(matrix,data):
    for inputLine in range (0,NUMBER_INPUT_LINES):
        addPair(1,2,data[inputLine],matrix)
        addPair(1,3,data[inputLine],matrix)
        addPair(1,4,data[inputLine],matrix)
        addPair(1,5,data[inputLine],matrix)
        addPair(1,6,data[inputLine],matrix)
        addPair(2,3,data[inputLine],matrix)
        addPair(2,4,data[inputLine],matrix)
        addPair(2,5,data[inputLine],matrix)
        addPair(2,6,data[inputLine],matrix)
        addPair(3,4,data[inputLine],matrix)
        addPair(3,5,data[inputLine],matrix)
        addPair(3,6,data[inputLine],matrix)
        addPair(4,5,data[inputLine],matrix)
        addPair(4,6,data[inputLine],matrix)
        addPair(5,6,data[inputLine],matrix)

def setMatrixPercentages(matrix): 
    for column in range (0,NUMBER_FEATURE_PAIRS):
        totalOccurences = matrix[4][column] + matrix[5][column]
        matrix[6][column] = matrix[4][column]/totalOccurences
        matrix[7][column] = matrix[5][column]/totalOccurences

    #print covarMatrix[6],covarMatrix[7]

def addPairPercentage(firstFeature,secondFeature,dataLine,matrix):
    firstValue = dataLine[firstFeature]
    secondValue = dataLine[secondFeature]
    column = getColumn(firstFeature,firstValue,secondFeature,secondValue,matrix)
    category0Value = matrix[6][column]
    category1Value = matrix[7][column]
    return [category0Value,category1Value]

def categoriseDataLine(dataLine,matrix):
    category0Value = 0
    category1Value = 0
    result = addPairPercentage(1,2,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(1,3,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(1,4,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(1,5,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(1,6,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(2,3,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(2,4,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(2,5,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(2,6,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(3,4,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(3,5,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(3,6,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(4,5,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(4,6,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    result = addPairPercentage(5,6,dataLine,matrix)
    category0Value += result[0]
    category1Value += result[1]

    #print category0Value,category1Value

    if (category0Value > category1Value):
        return 0
    elif (category0Value < category1Value):
        return 1
    else:
        print "draw",dataLine
        return -1

def categoriseData(data,matrix):
    right0 = 0
    wrong0 = 0
    right1 = 0
    wrong1 = 0
    draw = 0
    for inputLine in range (0,NUMBER_INPUT_LINES):
        predictedCategory = categoriseDataLine(data[inputLine],matrix)
        #print predictedCategory
        if (predictedCategory == -1):
            draw += 1   
        elif (predictedCategory == 0):
            if (data[inputLine][0] == 0):
                right0 +=1
            else:
                wrong0 +=1
        elif (predictedCategory == 1):
            if (data[inputLine][0] == 1):
                right1 +=1
            else:
                wrong1 +=1
        
    print right0,right1,wrong0,wrong1,draw


##--main---

data=readData("monkData/monks-1.train",NUMBER_INPUT_LINES)

covarMatrix = np.zeros(NUMBER_FEATURE_PAIRS*8)
covarMatrix = covarMatrix.reshape(8,NUMBER_FEATURE_PAIRS)

labelMatrix(covarMatrix)
fillMatrix(covarMatrix,data)
setMatrixPercentages(covarMatrix)

print "train"
categoriseData(data,covarMatrix)

print

testData=readData("monkData/monks-1.test",NUMBER_INPUT_LINES)
print "test"
categoriseData(testData,covarMatrix)


