"""
The monk 1 task is (f1=f2)or(f5=1)
This makes a net with three layers; the first connects to pairs, and both pairs
and first connect to output. Read in the pair and first to output.  Test
with just the input features.  The output to measure is in second.pkl.
"""
import sys
import cPickle as pickle

import pyNN.nest as sim

EXAMPLE_TIME = 50.0
NUMBER_CATEGORIES = 2
NUMBER_FEATURE_PAIRS = 121
PAIR_TIME = 50
TIME_BETWEEN_PAIR_INPUT_OUTPUT = 5

TIME_STEP = 1.0
FIRING_WEIGHT = 0.1
PAIR_WEIGHT = 0.05

NEURONS_PER_FEATURE = 1
NUMBER_INPUT_LINES = 124

RUN_DURATION = EXAMPLE_TIME*NUMBER_INPUT_LINES

neuronsInFirstLayer = 17
neuronsInSecondLayer = NUMBER_CATEGORIES

#initialize the simulator. 
def init(): 
    sim.setup(timestep=TIME_STEP,min_delay=TIME_STEP,
                    max_delay=TIME_STEP, debug=0)

def readData(fileName,inputLines):
    data = []
    dataFileHandle = open(fileName,'r')
    for lineNumber in range (0,NUMBER_INPUT_LINES):
        inputLine = dataFileHandle.readline()
        dataItemStrings = inputLine.split()
        category = int(dataItemStrings[0])
        feature0 = int(dataItemStrings[1])
        feature1 = int(dataItemStrings[2])
        feature2 = int(dataItemStrings[3])
        feature3 = int(dataItemStrings[4])
        feature4 = int(dataItemStrings[5])
        feature5 = int(dataItemStrings[6])
        dataLine = [category,feature0,feature1,feature2,feature3,feature4,feature5]
        #print dataLine,dataItemStrings[7]
        data = data + [dataLine]

    dataFileHandle.close()
    return data

#Make spike sources to turn on the first layer, one for each data line.
def createInputSpikeSources(start,numberSources):
    epochTime = EXAMPLE_TIME*numberSources
    inputSpikeTimesMatrix = []

    for sourceNumber in range (0,numberSources):
        inputSpikeTimes = []
        offset = sourceNumber * EXAMPLE_TIME
        inputSpikeTime = start+offset
        inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]

        inputSpikeTimesMatrix = inputSpikeTimesMatrix + [inputSpikeTimes]

    #print "inputSpikeTimes", inputSpikeTimesMatrix
    inputSpikeGens = sim.Population(numberSources,sim.SpikeSourceArray,
                {'spike_times': inputSpikeTimesMatrix},label="inputSpikes")

    return inputSpikeGens


#Make connections for the fistNeurons (all the feature values for the first
#feature) and all the rest of the neurons (pairwise) to the second order
#neuron.
def connectPairInputs(startFirst,lastFirst,lastRest,firstToNeuron):
    connector = []
    toNeuron = firstToNeuron
    #connect the first neurons and the rest in pairs to the spikeSources
    for firstNeuron in range(startFirst,lastFirst):
        for secondNeuron in range (lastFirst,lastRest):
            connector = connector + [(firstNeuron,toNeuron,PAIR_WEIGHT,
                        TIME_STEP)]
            connector = connector + [(secondNeuron,toNeuron,
                        PAIR_WEIGHT,TIME_STEP)]
            toNeuron += 1
            
    return connector

def connectFirstToPairs(firstCells,pairCells):
    inputPairConnector = connectPairInputs(0,3,neuronsInFirstLayer,0)
    inputPairConnector = inputPairConnector + connectPairInputs(3,6,
                    neuronsInFirstLayer,42)
    #print inputPairConnector
    inputPairConnector = inputPairConnector + connectPairInputs(6,8,
                    neuronsInFirstLayer,75)
    inputPairConnector = inputPairConnector + connectPairInputs(8,11,
                    neuronsInFirstLayer,93)
    inputPairConnector = inputPairConnector + connectPairInputs(11,15,
                    neuronsInFirstLayer,111)
    #print inputPairConnector

    inputPairFromListConnector = sim.FromListConnector(inputPairConnector)

    sim.Projection(firstCells,pairCells,inputPairFromListConnector, 
                   receptor_type='excitatory')


#This just creates the sources.  One input and one output for
#each data line.
def createTestSources(numberSources):
    firstSpike = 10
    inputSpikeGens = createInputSpikeSources (firstSpike,numberSources)

    return [inputSpikeGens]

def makeInputToFirstLayerSynapses(dataSourceNumber,dataItem):
    #dataItem[0] is the cat
    feature1Offset = (dataItem[1]-1)
    inputConnector =  [(dataSourceNumber,feature1Offset,FIRING_WEIGHT,TIME_STEP)]
    feature2Offset = (dataItem[2]-1) + 3 #There are 3 first features
    inputConnector =  inputConnector + [(dataSourceNumber,feature2Offset,FIRING_WEIGHT,TIME_STEP)]
    feature3Offset = (dataItem[3]-1) + 6 #There are 3 second features
    inputConnector =  inputConnector + [(dataSourceNumber,feature3Offset,FIRING_WEIGHT,TIME_STEP)]
    feature4Offset = (dataItem[4]-1) + 8 #There are 2 third features
    inputConnector =  inputConnector + [(dataSourceNumber,feature4Offset,FIRING_WEIGHT,TIME_STEP)]
    feature5Offset = (dataItem[5]-1) + 11 #There are 3 fourth features
    inputConnector =  inputConnector + [(dataSourceNumber,feature5Offset,FIRING_WEIGHT,TIME_STEP)]
    feature6Offset = (dataItem[6]-1) + 15 #There are 4 fifth features
    inputConnector =  inputConnector + [(dataSourceNumber,feature6Offset,FIRING_WEIGHT,TIME_STEP)]

    #print "source to input layer", inputConnector
    return inputConnector

def connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells,data):
    numberDataItems = NUMBER_INPUT_LINES
    for dataItem in range (0,numberDataItems):
        inputConnector= makeInputToFirstLayerSynapses(dataItem,data[dataItem])
        inputFromListConnector = sim.FromListConnector(inputConnector)
        #print "source to input",inputConnector
        sim.Projection(inputSpikeSources, inputCells, inputFromListConnector, 
                   receptor_type='excitatory')

def connectSpikeSourcesToSecondLayer(outputSpikeSources,outputCells):
    outputConnector = []
    primeWt = 0.05 #
    for lineNumber in range (0,NUMBER_INPUT_LINES):
        outputConnector = outputConnector + [(lineNumber,0,primeWt,TIME_STEP)]
        outputConnector = outputConnector + [(lineNumber,1,primeWt,TIME_STEP)]

            
    #print outputConnector
    outputFromListConnector = sim.FromListConnector(outputConnector)
    sim.Projection(outputSpikeSources, outputCells,outputFromListConnector, 
                   receptor_type='excitatory')

def connectInput(spikeSources, inputCells,outputCells,data):
    inputSpikeSources = spikeSources[0]

    connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells,data)
    connectSpikeSourcesToSecondLayer(inputSpikeSources,outputCells)

def readSynapses(fileName):
    connector = []

    fileHandle = open(fileName,'r')
    lines = fileHandle.readlines()
    for synapseNumber in range (0,len(lines)):
        line = lines[synapseNumber]
        fields = line.split(" ")
        fromNeuron = float(fields[0])
        toNeuron = float(fields[1])
        weight = float(fields[2])
        connector = connector + [(fromNeuron,toNeuron,weight,TIME_STEP)]
    fileHandle.close()

    return connector




def readInputToOutputLayer(fromCells, toCells):
    connector = []
    connector = readSynapses("output/firstSecondSynapses.txt")

    fromListConnector = sim.FromListConnector(connector)
    sim.Projection(fromCells,toCells,fromListConnector)


def readPairToOutputLayer(fromCells, toCells):
    connector = []
    connector = readSynapses("output/pairSecondSynapses.txt")

    fromListConnector = sim.FromListConnector(connector)
    sim.Projection(fromCells,toCells,fromListConnector)

def connectTestFirstToSecondLayer(fromCells, toCells, synapses):
    connector = []
    for synapseOffset in range (0,len(synapses)):
        fromNeuron = synapses[synapseOffset][0]
        toNeuron = synapses[synapseOffset][1]
        weight = synapses[synapseOffset][2]
        connector = connector + [(fromNeuron,toNeuron,
                        weight,TIME_STEP)]
    testFromListConnector = sim.FromListConnector(connector)
    sim.Projection(fromCells,toCells, testFromListConnector, 
                   receptor_type='excitatory')

def createDefaultNeurons(numberNeurons):
    cells=sim.Population(numberNeurons,sim.IF_cond_exp,cellparams={})
    return cells

def setupRecording(cells):
    cells.record({'spikes','v'})

def printResults(simCells,file):
    simCells.write_data(file+".pkl",'spikes')
    #simCells.write_data(file+"V.pkl",'v')
    #simCells.write_data(file+"W.pkl",'w')

def printF5Equal1(data):
    for lineNumber in range (0,NUMBER_INPUT_LINES):
        feature5 = data[lineNumber][5]
        if (feature5==1):
            print 62000+(lineNumber*50)

def printSTDPSynapses(fileName,synapseMatrix):
    #for synapse in synapseMatrix:
    #    print synapse[0],synapse[1],synapse[2]
    fileHandle = open(fileName,'w')
    for synapse in synapseMatrix:
        outString=str(synapse[0])+" "+ str(synapse[1])+" "+str(synapse[2])+"\n"
        fileHandle.write(outString)
    fileHandle.close()



##--main---
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 2):
    testDataFile = args[1]
else:
    testDataFile = "monkData/monks-1.test"
    #testDataFile = "monkData/monks-1.train"

simName = "nest"
spinnVersion = -1

init()
data=readData(testDataFile,NUMBER_INPUT_LINES)
testSpikeGenerators = createTestSources(NUMBER_INPUT_LINES)

firstLayerCells = createDefaultNeurons(neuronsInFirstLayer)
secondOrderLayerCells = createDefaultNeurons(NUMBER_FEATURE_PAIRS)
secondLayerCells = createDefaultNeurons(neuronsInSecondLayer)

#hook it up
connectInput(testSpikeGenerators,firstLayerCells,secondLayerCells,data)
connectFirstToPairs(firstLayerCells,secondOrderLayerCells)

#make the stdp synapses
readInputToOutputLayer(firstLayerCells,secondLayerCells)
readPairToOutputLayer(secondOrderLayerCells,secondLayerCells)

setupRecording(firstLayerCells)
setupRecording(secondOrderLayerCells)
setupRecording(secondLayerCells)

print RUN_DURATION
sim.run(RUN_DURATION)

printResults(firstLayerCells,"output/firstTest")
printResults(secondOrderLayerCells,"output/pairsTest")
printResults(secondLayerCells,"output/pairOutput")







