"""
This makes a two layer net, with spikesources for each of the four inputs
and each of the four output categories (blend them together in tesying).
The layers are connected by STDP synapses.
This makes an xor input output pair.  
It trains for NUMBER_TRAINING_EPOCHS.  The synapses are large for three
outputs from either input feature, with the best answer (from the timing
the largest) and the fourth much smaller.  The output neurons give the
right results for each of the four tests.
"""
import sys
import cPickle as pickle

import pyNN.nest as sim

TIME_BETWEEN_INPUT_OUTPUT = 5

EXAMPLE_TIME = 50.0
NUMBER_TEST_EPOCHS = 1
NUMBER_CATEGORIES = 4

TIME_STEP = 1.0
FIRING_WEIGHT = 0.1

NEURONS_PER_FEATURE = 1
NUMBER_FEATURES = 4

RUN_DURATION = EXAMPLE_TIME*NUMBER_FEATURES*(NUMBER_TEST_EPOCHS)

neuronsInFirstLayer = NEURONS_PER_FEATURE*NUMBER_FEATURES
neuronsInSecondLayer = NUMBER_CATEGORIES

firstSecond = -1 #global for printing out synaptic weights

#initialize the simulator. 
def init(): 
    sim.setup(timestep=TIME_STEP,min_delay=TIME_STEP,
                    max_delay=TIME_STEP, debug=0)

#Make 4 spike sources to turn on the first layer.
#One for each of the four types of input.
def createInputSpikeSources(start):
    numberDataItems = NUMBER_FEATURES

    epochTime = EXAMPLE_TIME*numberDataItems
    spikeSources = []
    inputSpikeTimesMatrix = []
    for featureValueNumber in range (0,numberDataItems):
        inputSpikeTimes = []
        if (featureValueNumber  == 0): #a 10,60
            offset = 0
            for epoch in range (0,NUMBER_TEST_EPOCHS): 
                inputSpikeTime = start+offset+ (epoch*epochTime)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + EXAMPLE_TIME
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        elif (featureValueNumber  == 1): #A 111 161
            for epoch in range (0,NUMBER_TEST_EPOCHS): 
                offset = 1
                inputSpikeTime = start+offset+(epoch*epochTime)+(EXAMPLE_TIME*2)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + EXAMPLE_TIME
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        elif (featureValueNumber  == 2): #b 12 112
            for epoch in range (0,NUMBER_TEST_EPOCHS): 
                offset = 2
                inputSpikeTime = start+offset+(epoch*epochTime)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + (EXAMPLE_TIME*2)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        elif (featureValueNumber  == 3): #B 64 164
            for epoch in range (0,NUMBER_TEST_EPOCHS): 
                offset = 4
                inputSpikeTime = start+offset+(epoch*epochTime)+EXAMPLE_TIME
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + (EXAMPLE_TIME*2)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        else: 
            print "error in createInputSS"

        inputSpikeTimesMatrix = inputSpikeTimesMatrix + [inputSpikeTimes]

    #print "inputSpikeTimes", inputSpikeTimesMatrix
    inputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': inputSpikeTimesMatrix},label="inputSpikes")

    return inputSpikeGens

def createTestSources():
    firstSpike = 10
    inputSpikeGens = createInputSpikeSources (firstSpike)

    return [inputSpikeGens]


def makeInputToFirstLayerSynapses(dataItem):
    inputConnector =  [(dataItem,dataItem,FIRING_WEIGHT,TIME_STEP)]

    #print "source to input layer", inputConnector
    return inputConnector

def connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells):
    numberDataItems = NUMBER_FEATURES
    for dataItem in range (0,numberDataItems):
        inputConnector= makeInputToFirstLayerSynapses(dataItem)
        inputFromListConnector = sim.FromListConnector(inputConnector)
        #print "source to input",inputConnector
        sim.Projection(inputSpikeSources, inputCells, inputFromListConnector, 
                   receptor_type='excitatory')

def connectInput(spikeSources, inputCells,outputCells):
    inputSpikeSources = spikeSources[0]

    connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells)


def readSynapses(fileName):
    connector = []

    fileHandle = open(fileName,'r')
    lines = fileHandle.readlines()
    for synapseNumber in range (0,len(lines)):
        line = lines[synapseNumber]
        fields = line.split(" ")
        fromNeuron = float(fields[0])
        toNeuron = float(fields[1])
        weight = float(fields[2])
        connector = connector + [(fromNeuron,toNeuron,weight,TIME_STEP)]
    fileHandle.close()

    return connector

def readFirstToSecondLayer(fileName,fromCells, toCells):
    connector = readSynapses(fileName)

    fromListConnector = sim.FromListConnector(connector)

    firstSecond = sim.Projection(fromCells,toCells,fromListConnector)

def connectTestFirstToSecondLayer(fromCells, toCells, synapses):
    connector = []
    for synapseOffset in range (0,len(synapses)):
        fromNeuron = synapses[synapseOffset][0]
        toNeuron = synapses[synapseOffset][1]
        weight = synapses[synapseOffset][2]
        connector = connector + [(fromNeuron,toNeuron,
                        weight,TIME_STEP)]
    testFromListConnector = sim.FromListConnector(connector)
    sim.Projection(fromCells,toCells, testFromListConnector, 
                   receptor_type='excitatory')

def createDefaultNeurons(numberNeurons):
    cells=sim.Population(numberNeurons,sim.IF_cond_exp,cellparams={})
    return cells

def setupRecording(cells):
    cells.record({'spikes','v'})

def printResults(simCells,file):
    simCells.write_data(file+".pkl",'spikes')
    #simCells.write_data(file+"V.pkl",'v')
    #simCells.write_data(file+"W.pkl",'w')

##--main---
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 3):
    x = 1
else:
    x = 2


simName = "nest"
spinnVersion = -1

init()
trainSpikeGenerators = createTestSources()

firstLayerCells = createDefaultNeurons(neuronsInFirstLayer)
secondLayerCells = createDefaultNeurons(neuronsInSecondLayer)

#hook it up
connectInput(trainSpikeGenerators,firstLayerCells,secondLayerCells)
readFirstToSecondLayer("sTDPSynapses.txt",firstLayerCells,secondLayerCells)

setupRecording(firstLayerCells)
setupRecording(secondLayerCells)

sim.run(RUN_DURATION)

printResults(firstLayerCells,"output/firstTest")
printResults(secondLayerCells,"output/secondTest")



