"""
This makes a two layer net, with spikesources for each of the four inputs
and each of the four output categories (blend them together in tesying).
The layers are connected by STDP synapses.
This makes an xor input output pair.  
It trains for NUMBER_TRAINING_EPOCHS.  The synapses are large for three
outputs from either input feature, with the best answer (from the timing
the largest) and the fourth much smaller.  The output neurons give the
right results for each of the four tests.
"""
import sys
import cPickle as pickle

import pyNN.nest as sim

TIME_BETWEEN_INPUT_OUTPUT = 5

EXAMPLE_TIME = 50.0
NUMBER_TRAINING_EPOCHS = 200
NUMBER_CATEGORIES = 4

TIME_STEP = 1.0
FIRING_WEIGHT = 0.1

NEURONS_PER_FEATURE = 1
NUMBER_FEATURES = 4

RUN_DURATION = EXAMPLE_TIME*NUMBER_FEATURES*(NUMBER_TRAINING_EPOCHS+1)

neuronsInFirstLayer = NEURONS_PER_FEATURE*NUMBER_FEATURES
neuronsInSecondLayer = NUMBER_CATEGORIES

firstSecond = -1 #global for printing out synaptic weights

#initialize the simulator. 
def init(): 
    sim.setup(timestep=TIME_STEP,min_delay=TIME_STEP,
                    max_delay=TIME_STEP, debug=0)

#Make 4 spike sources to turn on the first layer.
#One for each of the four types of input.
def createInputSpikeSources(start):
    numberDataItems = NUMBER_FEATURES

    epochTime = EXAMPLE_TIME*numberDataItems
    spikeSources = []
    inputSpikeTimesMatrix = []
    for featureValueNumber in range (0,numberDataItems):
        inputSpikeTimes = []
        if (featureValueNumber  == 0): #a 10,60
            offset = 0
            for epoch in range (0,NUMBER_TRAINING_EPOCHS+1): 
                inputSpikeTime = start+offset+ (epoch*epochTime)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + EXAMPLE_TIME
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        elif (featureValueNumber  == 1): #A 111 161
            for epoch in range (0,NUMBER_TRAINING_EPOCHS+1): 
                offset = 1
                inputSpikeTime = start+offset+(epoch*epochTime)+(EXAMPLE_TIME*2)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + EXAMPLE_TIME
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        elif (featureValueNumber  == 2): #b 12 112
            for epoch in range (0,NUMBER_TRAINING_EPOCHS+1): 
                offset = 2
                inputSpikeTime = start+offset+(epoch*epochTime)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + (EXAMPLE_TIME*2)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        elif (featureValueNumber  == 3): #B 64 164
            for epoch in range (0,NUMBER_TRAINING_EPOCHS+1): 
                offset = 4
                inputSpikeTime = start+offset+(epoch*epochTime)+EXAMPLE_TIME
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
                inputSpikeTime = inputSpikeTime + (EXAMPLE_TIME*2)
                inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
        else: 
            print "error in createInputSS"

        inputSpikeTimesMatrix = inputSpikeTimesMatrix + [inputSpikeTimes]

    #print "inputSpikeTimes", inputSpikeTimesMatrix
    inputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': inputSpikeTimesMatrix},label="inputSpikes")

    return inputSpikeGens

#Make 4 spike sources to turn on the second layer
def createOutputSpikeSources(start):
    #Create the output spike sources and when they fire
    epochTime = EXAMPLE_TIME*NUMBER_FEATURES
    outputSpikeTimesMatrix = []
    learnOffset = TIME_BETWEEN_INPUT_OUTPUT
    outputFalse1SpikeTimes = [] #f
    outputFalse2SpikeTimes = [] #t
    outputTrue1SpikeTimes = [] #T
    outputTrue2SpikeTimes = [] #F

    for epoch in range (0,NUMBER_TRAINING_EPOCHS): 
        outputSpikeTime = start+learnOffset+(epoch*epochTime)
        outputFalse1SpikeTimes = outputFalse1SpikeTimes + [outputSpikeTime]
        outputSpikeTime = start+learnOffset+(epoch*epochTime) + EXAMPLE_TIME
        outputTrue1SpikeTimes = outputTrue1SpikeTimes + [outputSpikeTime]
        outputSpikeTime = start+learnOffset+(epoch*epochTime) + (2*EXAMPLE_TIME)
        outputTrue2SpikeTimes = outputTrue2SpikeTimes + [outputSpikeTime]
        outputSpikeTime = start+learnOffset+(epoch*epochTime) + (3*EXAMPLE_TIME)
        outputFalse2SpikeTimes = outputFalse2SpikeTimes + [outputSpikeTime] 

    outputSpikeTimesMatrix = outputSpikeTimesMatrix + [outputFalse1SpikeTimes,
            outputTrue1SpikeTimes, outputTrue2SpikeTimes,outputFalse2SpikeTimes]

    outputSpikeGens = sim.Population(NUMBER_CATEGORIES,sim.SpikeSourceArray,
                {'spike_times': outputSpikeTimesMatrix},label="outputSpikes")

    #print "outputSpikeTimes", outputSpikeTimesMatrix
    return outputSpikeGens

def createTrainSources():
    firstSpike = 10
    inputSpikeGens = createInputSpikeSources (firstSpike)
    outputSpikeGens = createOutputSpikeSources(firstSpike)

    return [inputSpikeGens,outputSpikeGens]


def makeInputToFirstLayerSynapses(dataItem):
    inputConnector =  [(dataItem,dataItem,FIRING_WEIGHT,TIME_STEP)]

    #print "source to input layer", inputConnector
    return inputConnector

def connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells):
    numberDataItems = NUMBER_FEATURES
    for dataItem in range (0,numberDataItems):
        inputConnector= makeInputToFirstLayerSynapses(dataItem)
        inputFromListConnector = sim.FromListConnector(inputConnector)
        #print "source to input",inputConnector
        sim.Projection(inputSpikeSources, inputCells, inputFromListConnector, 
                   receptor_type='excitatory')

def connectSpikeSourcesToSecondLayer(outputSpikeSources,outputCells):
    outputConnector = []
    for category in range (0,NUMBER_CATEGORIES):
        outputConnector = outputConnector + [(category,category,FIRING_WEIGHT,TIME_STEP)]
            
    #print outputConnector
    outputFromListConnector = sim.FromListConnector(outputConnector)
    sim.Projection(outputSpikeSources, outputCells,outputFromListConnector, 
                   receptor_type='excitatory')

def connectInput(spikeSources, inputCells,outputCells):
    inputSpikeSources = spikeSources[0]
    outputSpikeSources = spikeSources[1]

    connectSpikeSourcesToFirstLayer(inputSpikeSources,inputCells)
    connectSpikeSourcesToSecondLayer(outputSpikeSources,outputCells)


def connectFirstToSecondLayer(fromCells, toCells):
    global firstSecond
    connector = []
    stdp_model = sim.STDPMechanism(
        #timing_dependence=sim.SpikePairRule(tau_plus=20.0,tau_minus=20.0,
        #                                    A_plus=0.005, A_minus=0.005),
        #weight_dependence=sim.MultiplicativeWeightDependence(w_min=0.0, w_max=0.03)#,
        timing_dependence=sim.SpikePairRule(tau_plus=20.0,tau_minus=20.0,
                                            A_plus=0.01, A_minus=0.01),
        weight_dependence=sim.MultiplicativeWeightDependence(w_min=0.0, w_max=0.06),
        weight=0.0000,
        delay=1.0,
        dendritic_delay_fraction=float(1))

    connector = []
    for firstLayerNeuron in range (0,neuronsInFirstLayer):
        for secondLayerNeuron in range (0, neuronsInSecondLayer):
            weight = 0.0
            connector = connector + [(firstLayerNeuron,
                    secondLayerNeuron,weight,TIME_STEP)]
    fromListConnector = sim.FromListConnector(connector)

    firstSecond = sim.Projection(fromCells,toCells,
                                 fromListConnector,
                                 synapse_type = stdp_model)

def connectTestFirstToSecondLayer(fromCells, toCells, synapses):
    connector = []
    for synapseOffset in range (0,len(synapses)):
        fromNeuron = synapses[synapseOffset][0]
        toNeuron = synapses[synapseOffset][1]
        weight = synapses[synapseOffset][2]
        connector = connector + [(fromNeuron,toNeuron,
                        weight,TIME_STEP)]
    testFromListConnector = sim.FromListConnector(connector)
    sim.Projection(fromCells,toCells, testFromListConnector, 
                   receptor_type='excitatory')

def createDefaultNeurons(numberNeurons):
    cells=sim.Population(numberNeurons,sim.IF_cond_exp,cellparams={})
    return cells

def setupRecording(cells):
    cells.record({'spikes','v'})

def printResults(simCells,file):
    simCells.write_data(file+".pkl",'spikes')
    #simCells.write_data(file+"V.pkl",'v')
    #simCells.write_data(file+"W.pkl",'w')

def printSTDPSynapses(fileName,synapseMatrix):
    #for synapse in synapseMatrix:
    #    print synapse[0],synapse[1],synapse[2]
    fileHandle = open(fileName,'w')
    for synapse in synapseMatrix:
        outString=str(synapse[0])+" "+ str(synapse[1])+" "+str(synapse[2])+"\n"
        fileHandle.write(outString)
    fileHandle.close()



##--main---
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 3):
    x = 1
else:
    x = 2


simName = "nest"
spinnVersion = -1

init()
trainSpikeGenerators = createTrainSources()

firstLayerCells = createDefaultNeurons(neuronsInFirstLayer)
secondLayerCells = createDefaultNeurons(neuronsInSecondLayer)

#hook it up
connectInput(trainSpikeGenerators,firstLayerCells,secondLayerCells)
connectFirstToSecondLayer(firstLayerCells,secondLayerCells)

setupRecording(firstLayerCells)
setupRecording(secondLayerCells)

print RUN_DURATION, EXAMPLE_TIME*NUMBER_FEATURES*NUMBER_TRAINING_EPOCHS
sim.run(RUN_DURATION)

printResults(firstLayerCells,"output/first")
printResults(secondLayerCells,"output/second")

firstSecondSynapseWeights = firstSecond.get(["weight"],format="list")
#print "stdp synapse matrix ", firstSecondSynapseWeights
#Print these to a file
printSTDPSynapses("sTDPSynapses.txt",firstSecondSynapseWeights)


